"""
WebDebug HTTP message

This is the superclass to both HTTPrequest and HTTPresponse.


WebDebug is web development proxy server that captures and displays HTTP
requests and responses.
Copyright (C) 1999  Paul Clip.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

"""

### IMPORTS

from WDconst import *

import string
import time
import math
import urllib


### CONSTANTS


### Classes

class HTTPmessage:
	"""
	The super class to HTTPrequest and HTTPresponse.
	Member variables:
		rawsize: the total size of the HTTP message in bytes, including headers
		creation_time: the time at which the message was instantiated
		ip_address: ip address of the remote host (browser for requests, server for responses)
	"""

	def __init__(self, ip):
		self.rawsize = 0
		self.creation_time = time.time()
		self.ip_address = ip

	def readLine(self, input):
		""" reads a line from input and keeps track of total size of message """
		str = input.readline()
		self.rawsize = self.rawsize + len(str)
		return str

	def getRawSize(self):
		""" Returns the raw size of the message in bytes """
		return self.rawsize

	def parseHeaders(self, input):
		""" parses HTTP message headers and returns two mappings
		    raw_headers come verbatim from the browser, headers have their names lowercased """
		headers = {}
		raw_headers = {}
		line = self.readLine(input)
		while ':' in line:
			i = string.find(line, ':')
			raw_headers[line[:i]] = string.strip(line[i+1:])
			headers[string.lower(line[:i])] = string.strip(line[i+1:])
			line = input.readline()
		return raw_headers, headers

	def readContents(self, input, read_anyway=0):
		""" reads the body of an HTTP message """
		raw_contents = ""
		content_length = self.headers.get('content-length')
		if content_length != None:
			raw_contents = input.read(int(content_length))
		elif read_anyway:
			# BUGBUG this may not read everything for very slow responses
			raw_contents = input.read()
		self.rawsize = self.rawsize + len(raw_contents)
		return raw_contents

	def findNameValue(self, str):
		""" return a url decoded name=value pair in the string str """
		equal = string.find(str, '=')
		if equal == -1:
			name = str
			value = ""
		else:
			name = urllib.unquote_plus(str[:equal])
			value = urllib.unquote_plus(str[equal+1:])
		return name, value
			
	def printError(self, errorstr):
		""" prints an error message to the console """
		print "HTTPmessage(%s) error: %s" % (id(self), errorstr)

	def stripTrailer(self, str):
		""" strips the tailing \\n or \\r\\n from a HTTP header line """
		requestline = str
		if requestline[-2:] == '\r\n':
			requestline = requestline[:-2]
		elif requestline[-1:] == '\n':
			requestline = requestline[:-1]
		return requestline

	def shortTime(self, t):
		""" return an HH:MM:SS.MS string based on time t """
		t1 = time.strftime("%H:%M:%S", time.localtime(t))
		return "%s.%s" % (t1, ("%.3f" % math.modf(t)[0])[2:])
	
	def toHTML(self, commandline):
		""" renders the common parts of the HTTP messages to HTML """
		html = "<table border=1 width=800><th align=left colspan=2 bgcolor=#dddddd>" + commandline + "</th>\n"
		for h in self.raw_headers.keys():
			html = html + "<tr><td bgcolor=#dddddd>%s:</td><td>%s</td>\n" % (h, self.raw_headers[h])
		html = html + "<tr><td>Size of message</td><td>%s bytes</td>\n" % self.getRawSize()
		html = html + "<tr><td>Message received</td><td>%s</td>\n" % self.shortTime(self.creation_time)
		return html

