"""
WebDebug HTTPrequest

This class wraps the HTTP proxy request.


WebDebug is web development proxy server that captures and displays HTTP
requests and responses.
Copyright (C) 1999  Paul Clip.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

"""

### IMPORTS

import HTTPmessage
from WDconst import *

import SocketServer
import re
import string
import base64


### CONSTANTS

DEFAULT_HTTP_PORT = 80

DEFAULT_HTTP_PATH = '/'

extractPathRE = re.compile('http://(.+?)(:.+?){0,1}?/(.+$){0,1}', re.I)


### Classes

class HTTPrequest(HTTPmessage.HTTPmessage):
	"""
	Encapsulates an HTTP request.
	Member variables:
		raw_requestline: raw first line of HTTP request
		requestline: raw_requestline minus trailing whitespace
		command: HTTP command (GET, POST, etc.)
		rawpath: the HTTP path that comes after the command on the first line of HTTP request
		version: the HTTP version of the request
		raw_headers: a mapping of the headers included in the request
		headers: a mapping of the headers included in the request (with names in lowercase)
		raw_contents: the raw content of the request, if any (typically found in POST requests)
		contents: a mapping of the parsed (non-URL decoded) raw_contents of the request, if any
		host: the host to connect to
		port: the port to use
		path: the path to request
	"""

	def __init__(self, input, ip):
		HTTPmessage.HTTPmessage.__init__(self, ip)
		self.raw_requestline = self.readLine(input)
		self.requestline = string.strip(self.raw_requestline)
		words = string.split(self.requestline)
		if len(words) == 3:
			[command, rawpath, version] = words
			if version[:5] != 'HTTP/':
				self.printError("Bad request version (%s)" % version)
				return
		else:
			self.printError("Bad request syntax (%s)" % self.requestline)
			return
		self.command, self.rawpath, self.version = command, rawpath, version
		self.host, self.port, self.path = self.parsePath()
		self.raw_headers, self.headers = self.parseHeaders(input)
		self.raw_contents = self.readContents(input)
		hasParams = string.find(self.path, '?')
		self.contents = {}
		if hasParams != -1:
			self.contents = self.parseContents(self.path[hasParams+1:])
		self.contents.update(self.parseContents(self.raw_contents))
		if self.isPOST():
			# POSTs seem to have a trailing newline. If we don't read it,
			# the browser flags an error
			self.readLine(input)

	def parsePath(self):
		""" parses the path given to WebDebug into its constituent host, port and path """
		# if it's not a proxy request
		if not startsWith(self.rawpath, 'http:'):
			return "", "", self.rawpath
		m = extractPathRE.match(self.rawpath)
		host, port, path = m.groups()
		if port == None:
			port = DEFAULT_HTTP_PORT
		elif port[0] == ':':
			port = string.atoi(port[1:])
		if path == None:
			path = DEFAULT_HTTP_PATH
		else:
			path = DEFAULT_HTTP_PATH + path
		return host, port, path

	def parseContents(self, raw_c):
		""" parse the name=value pairs included in the request, either on the URL itself, or from a POST """
		contents = {}
		if len(raw_c) != 0:
			while '&' in raw_c:
				sep = string.find(raw_c, '&')
				name, value = self.findNameValue(raw_c[:sep])
				contents[name] = value
				raw_c = raw_c[sep+1:]
			if len(raw_c) > 0:
				name, value = self.findNameValue(raw_c)
				contents[name] = value
		return contents

	def isPOST(self):
		""" returns true if the request is a POST """
		if string.lower(self.command) == 'post':
			return TRUE
		else:
			return FALSE

	def sendRequest(self, output):
		""" writes the request to output, proxy-related headers are not transmitted """
		output.write("%s %s %s\r\n" % (self.command, self.path, self.version))
		for h in self.raw_headers.keys():
			if string.lower(h)[:5] != 'proxy':
				output.write("%s: %s\r\n" % (h, self.raw_headers[h]))
		output.write("\r\n")
		output.write(self.raw_contents)
		output.flush()

	def getHost(self):
		return self.host

	def getPort(self):
		return self.port

	def getPath(self):
		return self.path

	def getContents(self):
		return self.contents

	def toHTML(self):
		""" renders the request to HTML """
		html = HTTPmessage.HTTPmessage.toHTML(self, self.raw_requestline)
		html = html + "<tr><td>Browser IP address</td><td>%s</td>\n" % self.ip_address
		# decode the basic authorization username and password if it's available
		if self.headers.get('authorization') != None:
			auth = base64.decodestring(string.split(self.headers.get('authorization'))[1])
			html = html + "<tr><td>Decoded authorization</td><td>%s</td>\n" % auth
		# display parameters, if any
		if len(self.contents) != 0:
			html = html + "<tr><td valign=\"top\">Parameters</td><td>"
			html = html + "<table border=0 cellspacing=\"2\"><tr><th align=left>Name</th><th>&nbsp;&nbsp;</th><th align=left>Value</th></tr>"
			for n in self.contents.keys():
				html = html + "<tr><td>%s</td><td>&nbsp;&nbsp;</td><td>%s</td></tr>\n" % (n, self.contents[n])
			html = html + "</table></td>\n"
		html = html + "</table>\n"
		return html
