"""
WebDebug Message List

Keeps track of all HTTP messages received by WebDebug, generates statistics for them
and renders them to HTML.


WebDebug is web development proxy server that captures and displays HTTP
requests and responses.
Copyright (C) 1999  Paul Clip.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

"""

### IMPORTS

from WDconst import *
import HTTPrequest
import HTTPresponse

import threading
import time


### CONSTANTS

# indexes for statistics gathering
STAT_NUMB = 0
STAT_LEN = 1

# indexes for saving and restoring state
MLIST = "messagelist"
FIRSTTIME = "firsttime"
LASTTIME = "lasttime"
VERSION = "version"


### Classes

class MessageList:
	"""
	Keeps track of HTTP requests and responses belonging to a single session.
	Member variables:
		messagelist: a (chronological) sequence of HTTP request/response pairs
		listlock: a lock protecting the messagelist from simultaneous access
		firstmessagetime: the time at which the first message pair was added
		lastmessagetime: the time at which the last message pair was added
	"""

	def __init__(self):
		self.messagelist = []
		self.listlock = threading.Lock()
		self.firstmessagetime = 0
		self.lastmessagetime = 0

	def __getstate__(self):
		""" returns the state to pickle """
		return { MLIST: self.messagelist,
			 FIRSTTIME: self.firstmessagetime,
			 LASTTIME: self.lastmessagetime,
			 VERSION: VERSION_NUMBER }

	def __setstate__(self, state):
		""" sets the state when unpickling """
		self.__init__()
		self.messagelist = state[MLIST]
		self.firstmessagetime = state[FIRSTTIME]
		self.lastmessagetime = state[LASTTIME]

	def addMessagePair(self, request, response):
		""" Add a request/response pair to the message list """
		pair = { REQUEST: request, RESPONSE: response }
		if len(self.messagelist) == 0:
			self.firstmessagetime = time.time()
		self.lastmessagetime = time.time()
		self.listlock.acquire()
		self.messagelist.append(pair)
		self.listlock.release()

	def clearList(self):
		""" Clears the message list of its contents """
		self.listlock.acquire()
		self.messagelist = []
		self.firstmessagetime = 0
		self.lastmessagetime = 0
		self.listlock.release()

	def findMessage(self, objectid):
		""" Returns the request or response matching the objectid """
		for p in self.messagelist:
			if id(p[REQUEST]) == objectid:
				return p[REQUEST]
			elif id(p[RESPONSE]) == objectid:
				return p[RESPONSE]
		return None

	def toHTML(self):
		""" Converts the contents of the list to HTML, suitable for display in a browser """
		if len(self.messagelist) == 0:
			return "<h3>Message List Empty</h3>"
		html = """
		<h3>Contents of Message List</h3><p>
		%s message pairs<br>First pair added %s<br>Last pair added %s<p>
		""" %  (len(self.messagelist),
			time.ctime(self.firstmessagetime),
			time.ctime(self.lastmessagetime))
		html = html + "<table border=0>"
		counter = 1
		totrawreqsize = 0
		totrawrespsize = 0
		statistics = { }
		# render in chronological order all message pairs to HTML
		for p in self.messagelist:
			html = html + """
			<tr><th align=left><font color=red>HTTP browser request %s</font></th></tr><tr><td>%s</td></tr>
			<tr><th align=left><font color=red>HTTP server response %s</font></th></tr><tr><td>%s</td></tr>
			<tr><td><hr></td></tr>
			<tr><td>&nbsp;</td></tr>
			""" %  (counter, p[REQUEST].toHTML(),
				counter, p[RESPONSE].toHTML())
			totrawreqsize = totrawreqsize + p[REQUEST].getRawSize()
			totrawrespsize = totrawrespsize + p[RESPONSE].getRawSize()
			counter = counter + 1
			self.calcStatistics(statistics, p[RESPONSE])
		html = html + """
		</table><p>
		<table border=0>
		<tr><td><b>Total request/response pairs:</b></td><td>%d</td></tr>
		<tr><td><b>Total size of requests:</b></td><td>%s bytes (average %d bytes/request)</td></tr>
		<tr><td><b>Total size of responses:</b></td><td>%s bytes (average %d bytes/response)</td></tr>
		<tr><td><b>Total size of both:</b></td><td>%s bytes</td></tr>
		</table><p>
		""" %  (len(self.messagelist),
			totrawreqsize, totrawreqsize/len(self.messagelist),
			totrawrespsize,  totrawrespsize/len(self.messagelist),
			totrawreqsize + totrawrespsize)
		html = html + self.statisticsToHTML(statistics)
		return html

	def statisticsToHTML(self, stats):
		""" generates the content-type statistics in HTML """
		html = "<table border=1><tr><th colspan=3>Content-type Statistics</th></tr>"
		html = html + "<tr><th>Content-type</th><th>Number</th><th>Total size (bytes)</th></tr>\n"
		keys = stats.keys()
		keys.sort()
		for c in keys:
			html = html + '<tr><td>%s</td><td align="right">%s</td><td align="right">%s</td></tr>\n' % (c, stats[c][STAT_NUMB], stats[c][STAT_LEN])
		html = html + "</table><p>"
		return html

	def calcStatistics(self, stats, response):
		""" Keeps track of content statistics """
		type = response.getContentType()
		length = response.getContentLength()
		if length == 0:
			return
		if type == None:
			self.addToStats(stats, "UNKNOWN", length)
		else:
			self.addToStats(stats, type, length)

	def addToStats(self, stats, type, length):
		if stats.get(type) == None:
			stats[type] = [1, length]
		else:
			stats[type][STAT_NUMB] = stats[type][STAT_NUMB] + 1
			stats[type][STAT_LEN] = stats[type][STAT_LEN] + length

myList = MessageList()
